/*
  dwOpenClient.cpp - MQTT dwOpen Client Library

*/

#include "dwArduinoWiFiOpenClient.h"
#include <string.h>

int isprint( char Character )
{
	if ( (Character > 31) && (Character < 126) )
		return 1;
	else
		return 0;
}

void hexdump(void *mem, unsigned int len)
{
	#define HEXDUMP_COLS 16
	char tmpBuf[10];
	
	unsigned int i, j;

	for(i = 0; i < len + ((len % HEXDUMP_COLS) ? (HEXDUMP_COLS - len % HEXDUMP_COLS) : 0); i++)
	{
			/* print offset */
			if(i % HEXDUMP_COLS == 0)
			{
				sprintf(tmpBuf, "0x%06X: ", i);
				Serial.print(tmpBuf);
			}

			/* print hex data */
			if(i < len)
			{
					sprintf(tmpBuf, "%02X ", 0xFF & ((char*)mem)[i]);
					Serial.print(tmpBuf);
			}
			else /* end of block, just aligning for ASCII dump */
			{
					sprintf(tmpBuf,"   ");
					Serial.print(tmpBuf);					
			}

			/* print ASCII dump */
			if(i % HEXDUMP_COLS == (HEXDUMP_COLS - 1))
			{
					for(j = i - (HEXDUMP_COLS - 1); j <= i; j++)
					{
							if(j >= len) /* end of block, not really printing */
							{
									Serial.print(' ');										
							}
							else if(isprint(((char*)mem)[j])) /* printable char */
							{
									//Serial.print(0xFF & ((char*)mem)[j]);
									sprintf(tmpBuf, "%c", ((char*)mem)[j]);
									Serial.print(tmpBuf);									
							}
							else /* other char */
							{
									Serial.print('.');
							}
					}
					Serial.println("");
			}
	}
}


dwOpenClient::dwOpenClient() {
   this->_client = NULL;
}

dwOpenClient::dwOpenClient(uint8_t *ip, uint16_t port, void (*callback)(char*,uint8_t*,unsigned int), WiFiClient *client) {
   this->_client = client;
   this->callback = callback;
   this->ip = ip;
   this->port = port;
   this->domain = NULL;
}

dwOpenClient::dwOpenClient(char* domain, uint16_t port, void (*callback)(char*,uint8_t*,unsigned int), WiFiClient *client) {
   this->_client = client;
   this->callback = callback;
   this->domain = domain;
   this->port = port;
}

boolean dwOpenClient::connect(char *id) {
   return connect(id,NULL,NULL,0,0,0,0);
}

boolean dwOpenClient::connect(char *id, char *user, char *pass) {
   return connect(id,user,pass,0,0,0,0);
}

boolean dwOpenClient::connect(char *id, char* willTopic, uint8_t willQos, uint8_t willRetain, char* willMessage)
{
   return connect(id,NULL,NULL,willTopic,willQos,willRetain,willMessage);
}

boolean dwOpenClient::connect(char *id, char *user, char *pass, char* willTopic, uint8_t willQos, uint8_t willRetain, char* willMessage) {
   if (!connected()) {
      int result = 0;

      if (domain != NULL) {
        result = _client->connect(this->domain, this->port);
      } else {
        result = _client->connect(this->ip, this->port);
      }

      if (result) {
         nextMsgId = 1;
         uint8_t d[10] = {0x00,0x06,'M','Q','I','s','d','p',MQTTPROTOCOLVERSION};
		 
         // Leave room in the buffer for header and variable length field
         uint16_t length = 5;
         unsigned int j;
         for (j = 0;j<9;j++) {
            buffer[length++] = d[j];
         }

         uint8_t v;
         if (willTopic) {
            v = 0x06|(willQos<<3)|(willRetain<<5);
         } else {
            v = 0x02;
         }

         if(user != NULL) {
            v = v|0x80;

            if(pass != NULL) {
               v = v|(0x80>>1);
            }
         }

         buffer[length++] = v;

         buffer[length++] = ((MQTT_KEEPALIVE) >> 8);
         buffer[length++] = ((MQTT_KEEPALIVE) & 0xFF);
         length = writeString(id,buffer,length);
         if (willTopic) {
            length = writeString(willTopic,buffer,length);
            length = writeString(willMessage,buffer,length);
         }

         if(user != NULL) {
            length = writeString(user,buffer,length);
            if(pass != NULL) {
               length = writeString(pass,buffer,length);
            }
         }

         write(MQTTCONNECT,buffer,length-5);

         lastInActivity = lastOutActivity = millis();

         while (!_client->available()) {
            unsigned long t = millis();
            if (t-lastInActivity > MQTT_KEEPALIVE*1000UL) {
               _client->stop();
               return false;
            }
         }
         uint16_t llen;
         uint16_t len = readPacket(&llen);

         if (len == 4 && buffer[3] == 0) {
            lastInActivity = millis();
            pingOutstanding = false;
            return true;
         }
      }
      _client->stop();
   }
   return false;
}

uint8_t dwOpenClient::readByte() {
   while(!_client->available()) {}
   return _client->read();
}

uint16_t dwOpenClient::readPacket(uint16_t* lengthLength)
{
   uint16_t len = 0;
   buffer[len++] = readByte();
   uint8_t multiplier = 1;
   uint16_t length = 0;
   uint8_t digit = 0;

   do
   {
      digit = readByte();
      buffer[len++] = digit;
      length += (digit & 127) * multiplier;
      multiplier *= 128;
   } while ((digit & 128) != 0);

   *lengthLength = len-1;

   for(uint16_t i = 0; i < length; i++)
   {
      if (len < MQTT_MAX_PACKET_SIZE)
      {
         buffer[len++] = readByte();
      }
      else
      {
         readByte();
         len = 0; // This will cause the packet to be ignored.
      }
   }

   //Serial.println("Recv:");
   //Serial.print("len=");
   //Serial.println(len);   
   //hexdump(buffer, len);
   
   return len;
}

boolean dwOpenClient::loop() {
   if (connected()) {

      unsigned long t = millis();
      if ((t - lastInActivity > MQTT_KEEPALIVE*1000UL) || (t - lastOutActivity > MQTT_KEEPALIVE*1000UL)) {
	  
         if (pingOutstanding) {
            _client->stop();
            return false;
         } else {
            buffer[0] = MQTTPINGREQ;
            buffer[1] = 0;
            _client->write(buffer,2);
            lastOutActivity = t;
            lastInActivity = t;
            pingOutstanding = true;
         }
      }

      if (_client->available()) {
	  		  
         uint16_t llen;
         uint16_t len = readPacket(&llen);
         if (len > 0) {
			 
            lastInActivity = t;
            uint8_t type = buffer[0]&0xF0;
            if (type == MQTTPUBLISH) {
				
               if (callback) {
				   
                  uint16_t tl = (buffer[llen+1]<<8)+buffer[llen+2];
                  char topic[tl+1];
                  for (uint16_t i=0;i<tl;i++) {
                     topic[i] = buffer[llen+3+i];
                  }
                  topic[tl] = 0;
                  // ignore msgID - only support QoS 0 subs
                  uint8_t *payload = buffer+llen+3+tl;
				  *(payload + len-llen-3-tl) = NULL;
                  callback(topic,payload,len-llen-3-tl);
               }
            } else if (type == MQTTPINGREQ) {
				
               buffer[0] = MQTTPINGRESP;
               buffer[1] = 0;
               _client->write(buffer,2);
            } else if (type == MQTTPINGRESP) {
               pingOutstanding = false;
            }
         }
      }
	  
      return true;
   }
   return false;
}

boolean dwOpenClient::publish(char* topic, char* payload) {
   return publish(topic,(uint8_t*)payload,strlen(payload),false);
}

boolean dwOpenClient::publish(char* topic, uint8_t* payload, unsigned int plength) {
   return publish(topic, payload, plength, false);
}

boolean dwOpenClient::publish(char* topic, uint8_t* payload, unsigned int plength, boolean retained) {
   if (connected()) {
      // Leave room in the buffer for header and variable length field
      uint16_t length = 5;
      length = writeString(topic,buffer,length);
      uint16_t i;
      for (i=0;i<plength;i++) {
         buffer[length++] = payload[i];
      }
      uint8_t header = MQTTPUBLISH;
      if (retained) {
         header |= 1;
      }
      return write(header,buffer,length-5);
   }
   return false;
}

boolean dwOpenClient::publish_P(char* topic, uint8_t* PROGMEM payload, unsigned int plength, boolean retained) {
   uint8_t llen = 0;
   uint8_t digit;
   int rc;
   uint16_t tlen;
   int pos = 0;
   int i;
   uint8_t header;
   unsigned int len;

   if (!connected()) {
      return false;
   }

   tlen = strlen(topic);

   header = MQTTPUBLISH;
   if (retained) {
      header |= 1;
   }
   buffer[pos++] = header;
   len = plength + 2 + tlen;
   do {
      digit = len % 128;
      len = len / 128;
      if (len > 0) {
         digit |= 0x80;
      }
      buffer[pos++] = digit;
      llen++;
   } while(len>0);

   pos = writeString(topic,buffer,pos);

   rc += _client->write(buffer,pos);

   for (i=0;i<plength;i++) {
      rc += _client->write((char)pgm_read_byte_near(payload + i));
   }

   lastOutActivity = millis();
   return rc == len + 1 + plength;
}

boolean dwOpenClient::write(uint8_t header, uint8_t* buf, uint16_t length) {
   uint8_t lenBuf[4];
   uint8_t llen = 0;
   uint8_t digit;
   uint8_t pos = 0;
   uint8_t rc;
   uint16_t len = length;
   do {
      digit = len % 128;
      len = len / 128;
      if (len > 0) {
         digit |= 0x80;
      }
      lenBuf[pos++] = digit;
      llen++;
   } while(len>0);

   buf[4-llen] = header;
   for (int i=0;i<llen;i++) {
      buf[5-llen+i] = lenBuf[i];
   }
   
   int msgLen = length+1+llen;
   
   memcpy( buffer, buf+(4-llen), msgLen);

   //Serial.println("Send:");   
   //hexdump(buffer, msgLen);
  
   int sendLen = msgLen;
   int offset = 0;
   unsigned char *WriteBuffer;
   while(msgLen)
   {   
	   if (msgLen > WIFI_SHIELD_WINDOW_SIZE)
	   {
		   sendLen = WIFI_SHIELD_WINDOW_SIZE;
		   WriteBuffer = &buffer[offset];
		   offset = offset + WIFI_SHIELD_WINDOW_SIZE;
	   }
	   else
	   {
		   WriteBuffer = &buffer[offset];		   
		   sendLen = msgLen;
	   }
	   
	   rc = _client->write(WriteBuffer, sendLen);    
	   
	   msgLen = msgLen - sendLen;
	}
	   
   lastOutActivity = millis();
   return (rc == sendLen);
}


boolean dwOpenClient::subscribe(char* topic) {
   if (connected()) {
	   
      // Leave room in the buffer for header and variable length field
      uint16_t length = 5;
      nextMsgId++;
      if (nextMsgId == 0) {
         nextMsgId = 1;
      }
      buffer[length++] = (nextMsgId >> 8);
      buffer[length++] = (nextMsgId & 0xFF);
      length = writeString(topic, buffer,length);
      buffer[length++] = 0; // Only do QoS 0 subs
      return write(MQTTSUBSCRIBE|MQTTQOS1,buffer,length-5);
   }
   return false;
}

boolean dwOpenClient::unsubscribe(char* topic) {
   if (connected()) {
      uint16_t length = 5;
      nextMsgId++;
      if (nextMsgId == 0) {
         nextMsgId = 1;
      }
      buffer[length++] = (nextMsgId >> 8);
      buffer[length++] = (nextMsgId & 0xFF);
      length = writeString(topic, buffer,length);
      return write(MQTTUNSUBSCRIBE|MQTTQOS1,buffer,length-5);
   }
   return false;
}

void dwOpenClient::disconnect() {
   buffer[0] = MQTTDISCONNECT;
   buffer[1] = 0;
   _client->write(buffer,2);
   _client->stop();
   lastInActivity = lastOutActivity = millis();
}

uint16_t dwOpenClient::writeString(char* string, uint8_t* buf, uint16_t pos) {
   char* idp = string;
   uint16_t i = 0;
   pos += 2;
   while (*idp) {
      buf[pos++] = *idp++;
      i++;
   }
   buf[pos-i-2] = (i >> 8);
   buf[pos-i-1] = (i & 0xFF);
   return pos;
}


boolean dwOpenClient::connected() {
   boolean rc;
   if (_client == NULL ) {
      rc = false;
   } else {
      rc = (int)_client->connected();
      if (!rc) _client->stop();
   }
   return rc;
}

// Generate JSON BindThing Request
void dwOpenClient::jsonCreateBindThing( char *thingKey, char *jsonText )
{
    // Create JSON Command Request...
    msgId++;
    sprintf(jsonText, "{ \"%d\": { \"command\": \"thing.bind\", \"params\": { \"key\": \"%s\" } } }", msgId, thingKey );

    return;
}

// Generate JSON UnbindThing Request
void dwOpenClient::jsonCreateUnbindThing( char *thingKey, char *jsonText )
{
    // Create JSON Command Request...
    msgId++;
    sprintf(jsonText, "{ \"%d\": { \"command\": \"thing.unbind\", \"params\": { \"key\": \"%s\" } } }", msgId, thingKey );

    return;
}

void dwOpenClient::jsonCreateLogPublish( char *thingKey, char *msgText, char *jsonText )
{
    // Create JSON Command Request...
    msgId++;
    sprintf(jsonText, "{ \"%d\": { \"command\": \"log.publish\", \"params\": {\"msg\": \"%s\", \"thingKey\": \"%s\" }}}", msgId, msgText, thingKey );

    return;
}

void dwOpenClient::jsonCreatePropertyPublish( char *thingKey, char *propKey, char *propValue, char *jsonText )
{
    // Create JSON Command Request...
    msgId++;
    sprintf(jsonText, "{ \"%d\": { \"command\": \"property.publish\", \"params\": {\"thingKey\": \"%s\",\"key\": \"%s\",\"value\": %s }}}", msgId, thingKey, propKey, propValue );

    return;
}

void dwOpenClient::jsonCreatePropertyCurrent( char *thingKey, char *propKey, char *jsonText )
{
    // Create JSON Command Request...
    msgId++;
    sprintf(jsonText, "{ \"%d\": { \"command\": \"property.current\", \"params\": {\"thingKey\": \"%s\",\"key\": \"%s\" }}}", msgId, thingKey, propKey );

    return;
}

void dwOpenClient::jsonCreateAlarmPublish( char *thingKey, char *alarmKey, int alarmState, char *alarmMsg, char *jsonText )
{
    // Create JSON Command Request...
    msgId++;
    sprintf(jsonText, "{ \"%d\": { \"command\": \"alarm.publish\", \"params\": {\"thingKey\": \"%s\", \"key\": \"%s\", \"msg\": \"%s\", \"state\": %d}}}", msgId, thingKey, alarmKey, alarmMsg, alarmState );

    return;
}

void dwOpenClient::jsonCreateLocationPublish( char *thingKey, char *locLatitude, char *locLongitude, char *locAltitude, char *locSpeed, char *locHeading, char *jsonText )
{
    // Create JSON Command Request...
    msgId++;
    sprintf(jsonText, "{ \"%d\": { \"command\": \"location.publish\", \"params\": {\"thingKey\": \"%s\", \"lat\": \"%s\", \"lng\": %s, \"altitude\": %s, \"speed\": %s, \"alti\": %s}}}", msgId, thingKey, locLatitude, locLongitude, locAltitude, locSpeed, locHeading );

    return;
}

void dwOpenClient::jsonCreateSetAttribute( char *thingKey, char *attribKey, char *attribValue, char *jsonText )
{
    // Create JSON Command Request...
    msgId++;
    sprintf(jsonText, "{ \"%d\": { \"command\": \"thing.attr.set\", \"params\": {\"thingKey\": \"%s\", \"key\": \"%s\", \"value\": \"%s\"}}}", msgId, thingKey, attribKey, attribValue );

    return;
}

void dwOpenClient::jsonCreateGetAttribute( char *thingKey, char *attribKey, char *jsonText )
{
    // Create JSON Command Request...
    msgId++;
    sprintf(jsonText, "{ \"%d\": { \"command\": \"thing.attr.get\", \"params\": {\"thingKey\": \"%s\", \"key\": \"%s\"}}}", msgId, thingKey, attribKey );

    return;
}

int dwOpenClient::dwLogPublish( char *thingKey, char *szLogMsg )
{
    // Publish Log Message...
    jsonCreateLogPublish( thingKey, szLogMsg, szBuffer );
    rc = publish( (char*) "api", szBuffer );

    delay(1000);
    loop();

    return rc;
}

int dwOpenClient::dwLocationPublish( char *thingKey, char *szLat, char *szLong, char *szAlt, char *szSpeed, char *szHeading )
{
    // Publish Location Message...
    jsonCreateLocationPublish( thingKey, szLat, szLong, szAlt, szSpeed, szHeading, szBuffer );
    rc = publish( (char*) "api", szBuffer );

    delay(1000);
    loop();

    return rc;
}

int dwOpenClient::dwPropertyPublish( char *thingKey, char *szKey, char *szValue )
{
    // Publish Property Message...
    jsonCreatePropertyPublish( thingKey, szKey, szValue, szBuffer );
    rc = publish( (char*) "api", szBuffer );

    delay(1000);
    loop();

    return rc;
}

int dwOpenClient::dwAttributePublish( char *thingKey, char *szKey, char *szValue )
{
    // Publish Attribute Message...
    jsonCreateSetAttribute( thingKey, szKey, szValue, szBuffer );
	
    rc = publish( (char*) "api", szBuffer );

    delay(1000);
    loop();

    return rc;
}

int dwOpenClient::dwAlarmPublish( char *thingKey, char *szKey, int iState, char *szMessage )
{
    // Publish Property Message...
    jsonCreateAlarmPublish( thingKey, szKey, iState, szMessage, szBuffer );
    rc = publish( (char*) "api", szBuffer );

    delay(1000);
    loop();

    return rc;
}

void dwOpenClient::dwSleep( int mSec )
{
  int loops = mSec / 1000;
  
  for( int i = 0; i < loops; i++ )
  {  
    delay(1000); loop();
  }
  
}


